import logging
import os
import unittest
import uuid
from configparser import ConfigParser
from pathlib import Path
from tempfile import NamedTemporaryFile

import boto3
import botocore
import pytest

from Pegasus import s3


class TestPaths(unittest.TestCase):
    def test_get_path_for_key(self):
        # s3.get_path_for_key(bucket, searchkey, key, output)
        self.assertEqual(s3.get_path_for_key("bucket", None, "foo", "baz"), "baz/foo")
        self.assertEqual(
            s3.get_path_for_key("bucket", None, "foo", "baz/"), "baz/bucket/foo"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", None, "foo/bar", "baz"), "baz/foo/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", None, "foo/bar", "baz/"), "baz/bucket/foo/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo", "foo/bar", "baz"), "baz/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo", "foo/bar", "baz/"), "baz/foo/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo", "foo/bar/boo", "baz"), "baz/bar/boo"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/", "foo/bar", "baz"), "baz/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/", "foo/bar", "baz/"), "baz/foo/bar"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/bar", "foo/bar", "baz"), "baz"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/bar", "foo/bar/boo", "baz"), "baz/boo"
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/bar", "foo/bar/boo", "baz/"),
            "baz/bar/boo",
        )
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo/bar", "foo/bar/boo/moo/choo", "baz/"),
            "baz/bar/boo/moo/choo",
        )
        self.assertEqual(s3.get_path_for_key("bucket", None, "foo/", "baz"), "baz/foo")
        self.assertEqual(s3.get_path_for_key("bucket", None, "foo", ""), "foo")
        self.assertEqual(s3.get_path_for_key("bucket", None, "foo/", ""), "foo")
        self.assertEqual(s3.get_path_for_key("bucket", None, "foo/bar", ""), "foo/bar")
        self.assertEqual(s3.get_path_for_key("bucket", "foo", "foo/", "baz"), "baz")
        self.assertEqual(s3.get_path_for_key("bucket", "foo/", "foo/", "baz"), "baz")
        self.assertEqual(
            s3.get_path_for_key("bucket", "foo", "foo/bar/", "baz"), "baz/bar"
        )

    def test_get_key_for_path(self):
        # s3.get_key_for_path(infile, path, outkey):
        self.assertEqual(s3.get_key_for_path("/foo", "/foo", "baz/"), "baz/foo")
        self.assertEqual(s3.get_key_for_path("/foo/", "/foo", "baz/"), "baz/foo")
        self.assertEqual(s3.get_key_for_path("/foo/", "/foo/", "baz/"), "baz/foo")
        self.assertEqual(s3.get_key_for_path("/foo", "/foo/", "baz/"), "baz/foo")
        self.assertEqual(s3.get_key_for_path("/foo", "/foo/bar", "baz"), "baz/bar")
        self.assertEqual(s3.get_key_for_path("/foo", "/foo/bar", "baz/"), "baz/foo/bar")
        self.assertEqual(
            s3.get_key_for_path("/foo", "/foo/bar/baz", "baz/"), "baz/foo/bar/baz"
        )
        self.assertEqual(
            s3.get_key_for_path("/foo", "/foo/bar/baz", "baz"), "baz/bar/baz"
        )
        self.assertEqual(
            s3.get_key_for_path("/foo", "/foo/bar/baz", "baz/boo/"),
            "baz/boo/foo/bar/baz",
        )
        self.assertRaises(Exception, s3.get_key_for_path, "/foo", "/foo/bar/baz", "")
        self.assertRaises(Exception, s3.get_key_for_path, "foo", "foo", "bar")
        self.assertRaises(Exception, s3.get_key_for_path, "/foo", "/bar", "bar")


# assumption is that .pegasus/credentials.conf exists with proper configuration
def is_missing_credentials():
    try:
        CFG_PATH = (Path.home() / ".pegasus/credentials.conf").resolve()
    except FileNotFoundError:
        return True

    cfg = ConfigParser()
    if len(cfg.read(str(CFG_PATH))) != 1:
        return True

    if "osgconnect" not in cfg and "rynge@osgconnect" not in cfg:
        return True

    return False


class TestConfig:
    def test_get_config_set_with_cli(self):
        with NamedTemporaryFile(mode="w+") as f:
            config = ConfigParser()
            config["DEFAULT"]["test"] = "some text"
            config.write(f)
            f.seek(0)

            parser, options = s3.parse_args(("-C", f.name))
            read_config = s3.get_config(options)
            assert read_config["DEFAULT"]["test"] == "some text"

    def test_get_config_set_with_env(self, monkeypatch):
        with NamedTemporaryFile(mode="w+") as f:
            config = ConfigParser()
            config["DEFAULT"]["test2"] = "some text"
            config.write(f)
            f.seek(0)

            monkeypatch.setenv("S3CFG", f.name)

            parser, options = s3.parse_args(())
            read_config = s3.get_config(options)
            assert read_config["DEFAULT"]["test2"] == "some text"

    @pytest.mark.skipif(
        is_missing_credentials(), reason="missing credentials, can't do test"
    )
    def test_get_config_from_default_location(self):
        credential_path = Path.home() / "credentials.conf"
        assert credential_path.exists()

        parser, options = s3.parse_args(())
        read_config = s3.get_config(options)

        assert "osgconnect" in read_config

    def test_get_config_from_old_default_location(self, monkeypatch):
        # set default credential path to a filename that should not exist in cwd
        monkeypatch.setattr(s3, "DEFAULT_CREDENTIAL_PATH", str(uuid.uuid1()))

        # create a temporary, secondary config
        with NamedTemporaryFile(mode="w+") as f:
            config = ConfigParser()
            config["DEFAULT"]["test3"] = "some text"
            config.write(f)
            f.seek(0)

            # overwrite the old default config location to be this temporary file
            monkeypatch.setattr(s3, "OLD_DEFAULT_CREDENTIAL_PATH", f.name)

            parser, options = s3.parse_args(())
            read_config = s3.get_config(options)

        assert "test3" in read_config["DEFAULT"]

    def test_config_not_found(self):
        parser, options = s3.parse_args(("-C", "bad_path_that_wont_exist087402170274x"))

        with pytest.raises(Exception) as e:
            s3.get_config(options)

        assert "Config file not found" in str(e)

    def test_config_has_invalid_permissions(self):
        with NamedTemporaryFile(mode="w") as f:
            # set the most liberal permissions for our fake config file
            parser, options = s3.parse_args(("-C", f.name))
            Path(f.name).chmod(0o777)
            with pytest.raises(Exception) as e:
                s3.get_config(options)

            assert "Permissions of config file {} are too liberal".format(
                f.name
            ) in str(e)


# --- testing pegasus-s3 commands ----------------------------------------------
@pytest.fixture(scope="module")
def s3_client():
    CFG_PATH = (Path.home() / ".pegasus/credentials.conf").resolve()
    os.environ["S3CFG"] = str(CFG_PATH)
    cfg = ConfigParser()
    if len(cfg.read(str(CFG_PATH))) != 1:
        raise RuntimeError(f"Could not find {CFG_PATH}")

    return boto3.client(
        "s3",
        endpoint_url=cfg["osgconnect"]["endpoint"],
        aws_access_key_id=cfg["rynge@osgconnect"]["access_key"],
        aws_secret_access_key=cfg["rynge@osgconnect"]["secret_key"],
    )


# for use in teardown functions as we cannot reuse the s3_client fixture there
def get_s3_client():
    CFG_PATH = (Path.home() / ".pegasus/credentials.conf").resolve()
    os.environ["S3CFG"] = str(CFG_PATH)
    cfg = ConfigParser()
    if len(cfg.read(str(CFG_PATH))) != 1:
        raise RuntimeError(f"Could not find {CFG_PATH}")

    return boto3.client(
        "s3",
        endpoint_url=cfg["osgconnect"]["endpoint"],
        aws_access_key_id=cfg["rynge@osgconnect"]["access_key"],
        aws_secret_access_key=cfg["rynge@osgconnect"]["secret_key"],
    )


@pytest.fixture(scope="function")
def test_file():
    test_file = Path("test_file")
    with test_file.open("w") as f:
        f.write("sample text")

    yield test_file

    test_file.unlink()


class TestLs:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_object(Bucket="test-bucket", Key="test_file")
            s3_client.delete_bucket(Bucket="test-bucket")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_ls_buckets(self, s3_client, capsys):
        resp = s3_client.list_buckets()
        expected_buckets = {b["Name"] for b in resp["Buckets"]}

        # pegasus-s3 ls s3://rynge@osgconnect
        parser, args = s3.parse_args(["ls", "s3://rynge@osgconnect"])
        s3.ls(args)

        # get stdout
        out, _ = capsys.readouterr()

        retrieved_buckets = {b.strip() for b in out.split()}
        assert expected_buckets == retrieved_buckets

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_ls_keys(self, s3_client, test_file, capsys):
        BUCKET = "test-bucket"

        # create test bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # add test file to bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # pegasus-s3 ls s3://rynge@osgconnect/test-bucket
        parser, args = s3.parse_args(["ls", "s3://rynge@osgconnect/test-bucket"])
        s3.ls(args)

        # get stdout
        out, _ = capsys.readouterr()

        # test
        assert out.strip() == test_file.name

        # cleanup bucket
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)


class TestCp:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_object(Bucket="test-bucket", Key="test_file")
            s3_client.delete_object(Bcuekt="test-bucket2", Key="test_file")
            s3_client.delete_bucket(Bucket="test-bucket")
            s3_client.delete_bucket(Bucket="test-bucket2")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_cp(self, s3_client, test_file):
        BUCKET = "test-bucket"
        BUCKET2 = "test-bucket2"

        # create test buckets
        s3_client.create_bucket(Bucket=BUCKET)
        s3_client.create_bucket(Bucket=BUCKET2)

        # add test file to bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # pegasus-s3 cp s3://rynge@osgconnect/test-bucket/test_file s3://rynge@osgconnect/test-bucket2/test_file
        parser, args = s3.parse_args(
            [
                "cp",
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
                f"s3://rynge@osgconnect/{BUCKET2}/{test_file.name}",
            ],
        )
        s3.cp(args)

        # ensure BUCKET2 has been created and test_file copied
        resp = s3_client.list_buckets()
        expected_buckets = {b["Name"] for b in resp["Buckets"]}

        assert BUCKET2 in expected_buckets

        try:
            s3_client.head_object(Bucket=BUCKET2, Key=test_file.name)
        except sbotocore.exceptions.ClientError:
            pytest.fail(
                "s3://rynge@osgconnect/{}/{} should exist but does not".format(
                    BUCKET2, test_file.name
                )
            )

        # clean up buckets
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)
        s3_client.delete_object(Bucket=BUCKET2, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET2)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_cp_force(self, s3_client, test_file):
        BUCKET = "test-bucket"
        BUCKET2 = "test-bucket2"

        # create test buckets
        s3_client.create_bucket(Bucket=BUCKET)
        s3_client.create_bucket(Bucket=BUCKET2)

        # add test file to buckets
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)
        s3_client.upload_file(str(test_file), BUCKET2, test_file.name)

        # pegasus-s3 cp --force s3://rynge@osgconnect/test-bucket/test_file s3://rynge@osgconnect/test-bucket2/test_file
        parser, args = s3.parse_args(
            [
                "cp",
                "--force",
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
                f"s3://rynge@osgconnect/{BUCKET2}/{test_file.name}",
            ],
        )
        s3.cp(args)

        # ensure test_file copied into bucket
        try:
            s3_client.head_object(Bucket=BUCKET2, Key=test_file.name)
        except sbotocore.exceptions.ClientError:
            pytest.fail(
                "s3://rynge@osgconnect/{}/{} should exist but does not".format(
                    BUCKET2, test_file.name
                )
            )

        # clean up buckets
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)
        s3_client.delete_object(Bucket=BUCKET2, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET2)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_cp_create_dest(self, s3_client, test_file):
        BUCKET = "test-bucket"
        BUCKET2 = "test-bucket2"

        # create test buckets
        s3_client.create_bucket(Bucket=BUCKET)

        # add test file to buckets
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # pegasus-s3 cp --create-dest s3://rynge@osgconnect/test-bucket/test_file s3://rynge@osgconnect/test-bucket2/test_file
        parser, args = s3.parse_args(
            [
                "cp",
                "--create-dest",
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
                f"s3://rynge@osgconnect/{BUCKET2}/{test_file.name}",
            ],
        )
        s3.cp(args)

        # ensure BUCKET2 has been created and test_file copied
        resp = s3_client.list_buckets()
        expected_buckets = {b["Name"] for b in resp["Buckets"]}

        assert BUCKET2 in expected_buckets

        try:
            s3_client.head_object(Bucket=BUCKET2, Key=test_file.name)
        except sbotocore.exceptions.ClientError:
            pytest.fail(
                "s3://rynge@osgconnect/{}/{} should exist but does not".format(
                    BUCKET2, test_file.name
                )
            )

        # clean up buckets
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)
        s3_client.delete_object(Bucket=BUCKET2, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET2)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_cp_create_dest_and_force(self, s3_client, test_file):
        BUCKET = "test-bucket"
        BUCKET2 = "test-bucket2"

        # create test bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # add test file to bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # pegasus-s3 cp --create-dest --force s3://rynge@osgconnect/test-bucket/test_file s3://rynge@osgconnect/test-bucket2/test_file
        parser, args = s3.parse_args(
            [
                "cp",
                "--create-dest",
                "--force",
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
                f"s3://rynge@osgconnect/{BUCKET2}/{test_file.name}",
            ],
        )
        s3.cp(args)

        # ensure BUCKET2 has been created an test_file copied
        resp = s3_client.list_buckets()
        expected_buckets = {b["Name"] for b in resp["Buckets"]}

        assert BUCKET2 in expected_buckets

        try:
            s3_client.head_object(Bucket=BUCKET2, Key=test_file.name)
        except sbotocore.exceptions.ClientError:
            pytest.fail(
                "s3://rynge@osgconnect/{}/{} should exist but does not".format(
                    BUCKET2, test_file.name
                )
            )

        # clean up buckets
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)
        s3_client.delete_object(Bucket=BUCKET2, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET2)


class TestMkdir:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_bucket(Bucket="test-bucket")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_mkdir(self, s3_client):
        BUCKET = "test-bucket"

        # pegasus-s3 mkdir s3://rynge@osgconnect/test-bucket
        parser, args = s3.parse_args(["mkdir", f"s3://rynge@osgconnect/{BUCKET}"])
        s3.mkdir(args)

        # ensure test-bucket has been created
        resp = s3_client.list_buckets()
        expected_buckets = {b["Name"] for b in resp["Buckets"]}
        assert BUCKET in expected_buckets

        # cleanup test-bucket
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_mkdir_bucket_already_owned_by_user(self, s3_client, caplog):
        BUCKET = "test-bucket2"

        # create test bucket before calling mkdir on same bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # pegasus-s3 mkdir s3://rynge@osgconnect/test-bucket
        parser, args = s3.parse_args(["mkdir", f"s3://rynge@osgconnect/{BUCKET}"])
        s3.mkdir(args)

        # check that log message printed
        assert caplog.record_tuples == [
            (
                "root",
                logging.WARNING,
                "Bucket: test-bucket2 exists and is already owned by user: rynge@osgconnect",
            )
        ]

        # cleanup test-bucket
        s3_client.delete_bucket(Bucket=BUCKET)


class TestRm:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_object(Bucket="test-bucket", Key="test_file")
            s3_client.delete_bucket(Bucket="test-bucket")

            s3_client.delete_object(Bucket="test-bucket1", Key="tb1_f1")
            s3_client.delete_object(Bucket="test-bucket1", Key="tb1_f2")
            s3_client.delete_bucket(Bucket="test-bucket1")

            s3_client.delete_object(Bucket="test-bucket2", Key="tb2_f1")
            s3_client.delete_object(Bucket="test-bucket2", Key="tb2_f2")
            s3_client.delete_bucket(Bucket="test-bucket2")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_rm(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # create test bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # add test file to bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # ensure file is there before test
        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            pytest.fail(f"Failed to upload test file: {test_file} to bucket: {BUCKET}")

        # remove file
        # pegasus-s3 rm s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            ["rm", f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}"]
        )
        s3.rm(args)

        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            # removed key and it should not exist
            pass

        # cleanup bucket
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_rm_force(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # create test bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # ensure file is NOT there before test
        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
            pytest.fail(f"File: {test_file.name} should not be in bucket: {BUCKET}")
        except botocore.exceptions.ClientError:
            pass

        # remove file
        # pegasus-s3 rm s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            ["rm", "--force", f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",]
        )
        s3.rm(args)

        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            # removed key and it should not exist
            pass

        # cleanup bucket
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_rm_using_file(self, s3_client):
        files = [Path("tb1_f1"), Path("tb1_f2"), Path("tb2_f1"), Path("tb2_f2")]

        # create test files
        for f in files:
            with f.open("w") as fp:
                fp.write(f.name)

        # mapping of keys to buckets
        buckets_keys = {"test-bucket1": files[0:2], "test-bucket2": files[2:]}

        # create test buckets and upload files
        for b, keys in buckets_keys.items():
            s3_client.create_bucket(Bucket=b)
            for k in keys:
                s3_client.upload_file(str(k), b, k.name)

        # create rm file
        rm_file = Path("rm_file")
        with rm_file.open("w") as fp:
            for b, keys in buckets_keys.items():
                for k in keys:
                    fp.write(f"s3://rynge@osgconnect/{b}/{k.name}\n")

        # remove file
        # pegasus-s3 rm s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(["rm", "--force", "--file", rm_file.name])
        s3.rm(args)

        # ensure keys were removed from bucket
        for b, keys in buckets_keys.items():
            for k in keys:
                try:
                    s3_client.head_object(Bucket=b, Key=k.name)
                    pytest.fail(f"Key: {k.name} should not exist in Bucket: {b}")
                except botocore.exceptions.ClientError:
                    pass

        # cleanup buckets
        for b, keys in buckets_keys.items():
            for k in keys:
                s3_client.delete_object(Bucket=b, Key=k.name)
            s3_client.delete_bucket(Bucket=b)

        # remove test files
        rm_file.unlink()
        for f in files:
            f.unlink()


class TestPut:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_object(Bucket="test-bucket", Key="test_file")
            s3_client.delete_bucket(Bucket="test-bucket")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_put(self, s3_client, test_file):
        BUCKET = "test-bucket-for-put-command"

        # create test bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # pegasus-s3 put test_file s3://rynge@osgconnect/test-bucket/test_file
        paser, args = s3.parse_args(
            [
                "put",
                str(test_file),
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
            ],
        )
        s3.put(args)

        # test
        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            pytest.fail(f"Key: {test_file.name} should exist in Bucket: {BUCKET}")

        # cleanup bucket
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_put_bucket_doesnt_exist(self, test_file):
        BUCKET = "test-bucket"

        # pegasus-s3 put test_file s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            ["put", str(test_file), f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",]
        )

        with pytest.raises(Exception) as e:
            s3.put(args)

        assert "Failed to upload file: test_file" in str(e)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_put_create_bucket(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # ensure test bucket doesn't yet exist
        try:
            s3_client.head_bucket(Bucket=BUCKET)
            pytest.fail(
                "test-bucket should not yet exist as it is to be created by the put command"
            )
        except botocore.exceptions.ClientError:
            pass

        # pegasus-s3 put --create-bucket test_file s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            [
                "put",
                "--create-bucket",
                str(test_file),
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
            ]
        )
        s3.put(args)

        # test
        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            pytest.fail(f"Key: {test_file.name} should exist in Bucket: {BUCKET}")

        # cleanup
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_put_key_already_exists(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # create bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # place file in bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # upload file (key that already exists in bucket)
        # pegasus-s3 put test_file s3://rynge@osgconnect/test-bucket/test_file
        paser, args = s3.parse_args(
            ["put", str(test_file), f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",]
        )

        with pytest.raises(Exception) as e:
            s3.put(args)

        assert "Key: test_file already exists" in str(e)

        # cleanup
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_put_force(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # create bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # place file in bucket so that pegasus-s3 put can overwrite it
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # upload file
        # pegasus-s3 put --force test_file s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            [
                "put",
                "--force",
                str(test_file),
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
            ]
        )
        s3.put(args)

        try:
            s3_client.head_object(Bucket=BUCKET, Key=test_file.name)
        except botocore.exceptions.ClientError:
            pytest.fail(f"Key: {test_file.name} should exist in Bucket: {BUCKET}")

        # cleanup
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)


class TestGet:
    def teardown_method(self, method):
        s3_client = get_s3_client()

        # cleanup in case test fails
        try:
            s3_client.delete_object(Bucket="test-bucket", Key="test_file")
            s3_client.delete_bucket(Bucket="test-bucket")
        except s3_client.exceptions.NoSuchBucket:
            pass

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_get(self, s3_client):
        BUCKET = "test-bucket"

        test_file = Path("test_file")
        with test_file.open("w") as f:
            f.write("sample text\n")

        # create bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # place file in bucket so that pegasus-s3 put can overwrite it
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # remove file
        test_file.unlink()

        # get file
        # pegasus-s3 get s3://rynge@osgconnect/test-bucket/test_file
        parser, args = s3.parse_args(
            ["get", f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}"]
        )
        s3.get(args)

        # file should exist again
        assert test_file.exists() == True
        test_file.unlink()

        # cleanup
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)

    @pytest.mark.skipif(is_missing_credentials(), reason="missing credentials")
    def test_get_as_file(self, s3_client, test_file):
        BUCKET = "test-bucket"

        # create bucket
        s3_client.create_bucket(Bucket=BUCKET)

        # place file in bucket
        s3_client.upload_file(str(test_file), BUCKET, test_file.name)

        # get file
        # pegasus-s3 get s3://rynge@osgconnect/test-bucket/test_file new_test_file
        parser, args = s3.parse_args(
            [
                "get",
                f"s3://rynge@osgconnect/{BUCKET}/{test_file.name}",
                "new_test_file",
            ]
        )
        s3.get(args)

        # check that test_file downloaded as new_test_file
        new_test_file = Path("new_test_file")
        assert new_test_file.exists() == True

        # cleanup
        new_test_file.unlink()
        s3_client.delete_object(Bucket=BUCKET, Key=test_file.name)
        s3_client.delete_bucket(Bucket=BUCKET)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import argparse
import importlib
import logging
import logging.handlers
import os
import re
import signal
import sys
import tarfile
import threading
import time
from pathlib import Path
from typing import Iterable, List, Set

ptransfer_module = importlib.import_module("Pegasus.cli.pegasus-transfer")
pegasus_transfer = ptransfer_module.pegasus_transfer

log = logging.getLogger("pegasus-checkpoint")

# name of archived & compressed checkpoints
CHECKPOINT_FILENAME = "pegasus.checkpoint.tar.gz"

# pegasus-transfer url file expected to be generated by pegasus-lite
PEGASUS_TRANSFER_URL_FILE = "pegasus_checkpoint_transfer_urls.json"

# file to which this process's PID will be written upon startup
PEGASUS_CHECKPOINT_PID_FILE = "pegasus_checkpoint.pid"


def parse_args(args: List[str] = sys.argv[1:]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Periodically transfer checkpoints back to staging site."
    )

    parser.add_argument(
        "-p",
        "--pattern",
        action="append",
        dest="patterns",
        metavar="PATTERN",
        help="""regex patterns to match when searching for files to checkpoint""",
    )

    def validate_interval(value: str) -> int:
        """
        Ensure that given interval is an integer and is nonzero.

        :param value: the given interval to validate
        :type value: str
        :raises argparse.ArgumentTypeError: interval cannot be casted to an int or is a negative integer
        :return: the given value
        :rtype: int
        """
        is_valid = True
        try:
            value = int(value)
        except ValueError:
            is_valid = False

        if not is_valid or value < 0:
            raise argparse.ArgumentTypeError(
                f"interval {value} must be a nonnegative integer"
            )

        return value

    parser.add_argument(
        "-i",
        "--interval",
        type=validate_interval,
        metavar="SECONDS",
        help="""interval in seconds at which pegasus-checkpoint will send
                checkpoint file back to the staging site
                """,
    )

    # TODO: add --once to indicate pegasus-checkpoint is to run CheckpointWorker.run once

    parser.add_argument(
        "-d", "--debug", action="store_true", default=False, help="enable debug logging"
    )

    parser.add_argument(
        "-l",
        "--log-to-file",
        action="store_true",
        default=False,
        help="enable logging to file pegasus-checkpoint.log",
    )

    return parser.parse_args(args)


def configure_logging(level: int, log_to_file: bool):
    """
    Setup logging.

    :param level: log level to use; set to :code:`logging.DEBUG` to enable debug logging
    :type level: int
    :param log_to_file: whether or not to write logs to the file pegasus-checkpoint.log
    :type log_to_file: bool
    """
    if level == logging.DEBUG:
        level = logging.DEBUG
    else:
        level = logging.INFO

    handlers = [logging.StreamHandler()]
    if log_to_file:
        handlers.append(
            logging.handlers.RotatingFileHandler(
                filename="pegasus-checkpoint.log",
                mode="a",
                maxBytes=(1 << 20),
                backupCount=2,
            )
        )

    logging.basicConfig(
        level=level,
        format="%(asctime)s %(threadName)s [%(levelname)s] %(message)s",
        handlers=handlers,
    )


def write_pid():
    """Write PID to :code:`PEGASUS_CHECKPOINT_PID_FILE`"""
    with open(PEGASUS_CHECKPOINT_PID_FILE, "w") as f:
        pid = os.getpid()
        f.write(str(pid))

        log.debug(f"started with PID: {pid}")


class PeriodicCheckpointNotifier(threading.Thread):
    """
    Thread to periodically signal to the worker thread to start saving and
    staging out checkpoint files.
    """

    log = logging.getLogger("pegasus-checkpoint.PeriodicCheckcpointNotifier")

    def __init__(self, interval: int, notify: threading.Event):
        """Constructor

        :param interval: interval in seconds to sleep for
        :type interval: int
        :param notify: event to set, which will notify checkpoint worker thread to wake up and run 
        :type notify: threading.Event
        """
        super().__init__(group=None, target=None, name="PeriodicCheckpointNotifier")
        self.interval = interval
        self.notify = notify

    def run(self):
        """
        Repeatedly sleep for the given interval, then notify checkpoint worker
        thread to start working upon waking up.
        """
        self.log.debug(f"{self.name} started with interval: {self.interval} seconds")
        while True:
            time.sleep(self.interval)
            self.notify.set()
            self.log.debug("notify event set")


class CheckpointWorker(threading.Thread):
    """Thread to handle saving and staging out checkpoint files."""

    log = logging.getLogger("pegasus-checkpoint.CheckpointWorker")

    def __init__(self, notify: threading.Event, patterns: List[str]):
        """Constructor

        :param notify: event to watch for, which will indicate that this thread needs to run
        :type notify: threading.Event
        :param patterns: regex file patterns to match
        :type patterns: List[str]
        """
        super().__init__(group=None, target=None, name="CheckpointWorker")

        self.notify = notify
        self.patterns = patterns

    def run(self):
        """
        Wait for notification to start working. Once notified, create the
        checkpoint archive file and transfer back to staging site with either
        pegasus-transfer or condor_chirp.
        self.log.debug("{} started".format(self.name))
        """
        while True:
            # wait till we are notified
            self.notify.wait(timeout=None)
            self.log.debug("got notification, starting work")

            # work
            matched_files = CheckpointWorker.get_matched_filenames(self.patterns)

            CheckpointWorker.archive_and_compress(matched_files)

            # TODO: add exception handling
            pegasus_transfer(
                max_attempts=3,
                num_threads=8,
                file=PEGASUS_TRANSFER_URL_FILE,
                symlink=True,
            )

            # clear the notification, so we can wait until it is set again
            self.notify.clear()
            self.log.debug("work done, cleared notification")

    @staticmethod
    def get_matched_filenames(patterns: List[str]) -> Set[str]:
        """
        Get all files/folders in the current working directory that match the
        given file patterns.

        :param patterns: regex patterns to match filenames with
        :type patterns: List[str]
        :return: all files and or folders which match one of the given patterns
        :rtype: Set[str]
        """
        patterns = [re.compile(p) for p in patterns]
        files_in_cwd = [str(f) for f in Path(".").iterdir()]
        matched = set()

        # n^2.. maybe can do a little better by pulling out files from files_in_cwd (using a set
        # isntead) as we make matches so we don't try to iterate over it again
        for p in patterns:
            for f in files_in_cwd:
                is_match = bool(p.fullmatch(f))

                if is_match:
                    matched.add(f)

                CheckpointWorker.log.debug(
                    "pattern: {:<30}, file: {:>20}, match: {}".format(
                        p.pattern, f, is_match
                    )
                )

        CheckpointWorker.log.info(
            f"given patterns matched the following filenames: {matched}"
        )

        return matched

    @staticmethod
    def archive_and_compress(filenames: Iterable[str]):
        """
        Given a list of file/folder names in CWD, archive and compress them into a single file
        :code:`pegasus.checkpoint.tar.gz`

        :param filenames: list of filenames to archive and compress 
        :type filenames: Iterable[str]
        """
        start = time.perf_counter()
        with tarfile.open(CHECKPOINT_FILENAME, "w|gz") as tar:
            for f in filenames:
                tar.add(name=f, recursive=True)
        end = time.perf_counter()

        checkpoint_size_in_mb = Path(CHECKPOINT_FILENAME).stat().st_size / (1 << 20)

        CheckpointWorker.log.info(
            "created {} MB checkpoint file: {}".format(
                checkpoint_size_in_mb, CHECKPOINT_FILENAME
            )
        )
        CheckpointWorker.log.info(f"archive and gzip took {end - start} seconds")


if __name__ == "__main__":
    args = parse_args()

    log_level = logging.INFO
    if args.debug:
        log_level = logging.DEBUG

    configure_logging(level=log_level, log_to_file=args.log_to_file)

    write_pid()

    # event to signal to worker thread to start saving checkpoints
    notify = threading.Event()

    # main thread to handle signal from user application
    def SIGUSR1_handler(signum, frame):
        log.info("SIGUSR1 received, setting notification")
        notify.set()

    signal.signal(signal.SIGUSR1, SIGUSR1_handler)

    # create and start a periodic checkpoint notifier if user specified an interval to run on
    periodic_notifier = None
    if args.interval:
        periodic_notifier = PeriodicCheckpointNotifier(
            interval=args.interval, notify=notify
        )
        periodic_notifier.start()

    worker = CheckpointWorker(notify=notify, patterns=args.patterns)
    worker.start()

    if periodic_notifier:
        periodic_notifier.join()
    worker.join()



import configparser
import io
import json
import logging
import os
import stat
import sys
import tempfile
from pathlib import Path

import pytest

from Pegasus import transfer


class TestReadJSONFormat:
    def test_read_json_format(self):
        data = [
            {
                "type": "transfer",
                "id": 1,
                "src_urls": [{"site_label": "web", "url": "http://pegasus.isi.edu"}],
                "dest_urls": [{"site_label": "local", "url": "file:///tmp/index.html"}],
            }
        ]

        data_str = json.dumps(data)
        inputs_loaded = list()
        transfer.read_json_format(input=data_str, inputs_l=inputs_loaded)

        expected_transfer_obj = transfer.Transfer()
        expected_transfer_obj._src_urls = [
            transfer.PegasusURL(
                url="http://pegasus.isi.edu", file_type="x", site_label="web"
            )
        ]
        expected_transfer_obj._dst_urls = [
            transfer.PegasusURL(
                url="file:///tmp/index.html", file_type="x", site_label="local"
            )
        ]

        assert inputs_loaded[0] == expected_transfer_obj

    def test_read_json_format_parse_error(self):
        data = [
            {
                "type": "transfer",
                "id": 1,
                "src_urls": [{"site_lael": "web", "url": "http://pegasus.isi.edu"}],
                "dest_urls": [{"site_label": "local", "url": "file:///tmp/index.html"}],
            }
        ]

        data_str = json.dumps(data)
        inputs_loaded = list()
        with pytest.raises(RuntimeError) as e:
            transfer.read_json_format(input=data_str, inputs_l=inputs_loaded)

        assert "Error parsing the transfer" in str(e)

    def test_read_json_format_unknown_entry_error(self):
        data = [
            {
                "type": "badtype",
                "id": 1,
                "src_urls": [{"site_lael": "web", "url": "http://pegasus.isi.edu"}],
                "dest_urls": [{"site_label": "local", "url": "file:///tmp/index.html"}],
            }
        ]

        data_str = json.dumps(data)
        inputs_loaded = list()
        with pytest.raises(RuntimeError) as e:
            transfer.read_json_format(input=data_str, inputs_l=inputs_loaded)

        assert "Unknown JSON entry:" in str(e)


@pytest.fixture
def cleanup_pegasus_credentials_env():
    """If PEGASUS_CREDENTIALS was set, remove it from the environment"""
    yield
    if os.environ["PEGASUS_CREDENTIALS"] != None:
        del os.environ["PEGASUS_CREDENTIALS"]


class TestLoadCredentials:
    def test_load_credentials(self, cleanup_pegasus_credentials_env):
        with tempfile.NamedTemporaryFile(mode="w+", delete=True) as f:
            creds = """
            [amazon]
            endpoint = https://s3.amazonaws.com/

            [joe@amazon]
            access_key = 99001122
            secret_key = abababababababababababababababab
            """
            f.write(creds)
            f.seek(0)

            os.environ["PEGASUS_CREDENTIALS"] = f.name

            credentials = transfer.load_credentials()
            assert credentials["amazon"]["endpoint"] == "https://s3.amazonaws.com/"
            assert credentials["joe@amazon"]["access_key"] == "99001122"
            assert (
                credentials["joe@amazon"]["secret_key"]
                == "abababababababababababababababab"
            )

    def test_loading_credentials_file_that_doesnt_exist(
        self, cleanup_pegasus_credentials_env
    ):
        os.environ["PEGASUS_CREDENTIALS"] = "bad_credentials_file..."
        with pytest.raises(RuntimeError) as e:
            transfer.load_credentials()

        assert "Credentials file does not exist" in str(e)

    def test_invalid_credentials_file_permissions(
        self, cleanup_pegasus_credentials_env
    ):
        with tempfile.NamedTemporaryFile(mode="w+", delete=True) as f:
            os.environ["PEGASUS_CREDENTIALS"] = f.name
            f.write("stuff")

            # change permissions to one that would cause load_credentials to
            # raise an exception
            os.chmod(path=f.name, mode=stat.S_IRWXG)

            with pytest.raises(RuntimeError) as e:
                transfer.load_credentials()

            assert "Permissions of credentials file" in str(e)

    def test_unable_to_read_credentials(self, caplog, cleanup_pegasus_credentials_env):
        caplog.set_level(logging.CRITICAL, logger="Pegasus")

        with tempfile.NamedTemporaryFile(mode="w+", delete=True) as f:
            f.write("x;kwj0923fjalksdjf438hfjojTHIS_IS_A_BAD_CONFIG_ldkfjals")
            f.seek(0)

            os.environ["PEGASUS_CREDENTIALS"] = f.name

            with pytest.raises(configparser.MissingSectionHeaderError) as e:
                transfer.load_credentials()

            assert "Unable to load credentials" in str(caplog.record_tuples)


class TestPegasusTransferInvocation:
    def test_error_opening_input_file(self):
        with pytest.raises(FileNotFoundError) as e:
            transfer.pegasus_transfer(
                max_attempts=3, num_threads=8, file="badfile", symlink=True
            )

    def test_transfers_read_from_stdin_succeeded(self, caplog):
        caplog.set_level(logging.INFO, logger="Pegasus")
        with tempfile.TemporaryDirectory() as td:
            temp_dir = Path(td)

            downloaded_files = list()
            xfers = list()
            NUM_XFERS = 5
            for i in range(NUM_XFERS):
                xfers.append(
                    {
                        "type": "transfer",
                        "id": i,
                        "src_urls": [
                            {"site_label": "web", "url": "http://pegasus.isi.edu"}
                        ],
                        "dest_urls": [
                            {
                                "site_label": "local",
                                "url": "file://" + str(temp_dir / f"index{i}.html"),
                            }
                        ],
                    }
                )
                downloaded_files.append(f"index{i}.html")

            sys.stdin = io.StringIO(json.dumps(xfers))

            is_successful = transfer.pegasus_transfer(
                max_attempts=1, num_threads=8, file=None, symlink=True
            )

            assert is_successful == True

            logs = str(caplog.record_tuples)
            assert f"{NUM_XFERS} transfers loaded" in logs
            assert "All transfers completed" in logs

            for f in downloaded_files:
                assert (temp_dir / f).exists()

    def test_transfers_read_from_file_succeeded(self, caplog):
        caplog.set_level(logging.DEBUG, logger="Pegasus")
        with tempfile.TemporaryDirectory() as td:
            temp_dir = Path(td)

            downloaded_files = list()
            xfers = list()
            NUM_XFERS = 8
            for i in range(NUM_XFERS):
                xfers.append(
                    {
                        "type": "transfer",
                        "id": i,
                        "src_urls": [
                            {"site_label": "web", "url": "http://pegasus.isi.edu"}
                        ],
                        "dest_urls": [
                            {
                                "site_label": "local",
                                "url": "file://" + str(temp_dir / f"index{i}.html"),
                            }
                        ],
                    }
                )
                downloaded_files.append(f"index{i}.html")

            with tempfile.NamedTemporaryFile("w+", delete=True) as f:
                f.write(json.dumps(xfers))
                f.seek(0)

                is_successful = transfer.pegasus_transfer(
                    max_attempts=1, num_threads=8, file=f.name, symlink=True
                )

            assert is_successful == True

            logs = str(caplog.record_tuples)
            assert f"{NUM_XFERS} transfers loaded" in logs
            assert f"Using {8} threads for this set of transfers" in logs
            assert "All transfers completed" in logs

            for f in downloaded_files:
                assert (temp_dir / f).exists()

    def test_some_transfers_failed(self, caplog):
        caplog.set_level(logging.CRITICAL, logger="Pegasus")

        with tempfile.NamedTemporaryFile(mode="w+", delete=True) as f:
            data = [
                {
                    "type": "transfer",
                    "id": 1,
                    "src_urls": [
                        {
                            "site_label": "web",
                            "url": "http://BADURL0983012830181028310831adna02a0o.net",
                        }
                    ],
                    "dest_urls": [
                        {"site_label": "local", "url": "file:///tmp/index.html"}
                    ],
                }
            ]

            f.write(json.dumps(data))
            f.seek(0)

            is_successful = transfer.pegasus_transfer(
                max_attempts=1, num_threads=8, file=f.name, symlink=True
            )

            assert is_successful == False
            assert "Some transfers failed!" in str(caplog.record_tuples)



import os
import subprocess
import sys

import pytest

from Pegasus import cli

PYTHON_INTERPRETER = sys.executable
PEGASUS_INTEGRITY = cli.__path__[0] + "/pegasus-integrity.py"


@pytest.fixture(scope="package")
def data_files(resource_path_root):
    cwd = os.getcwd()
    os.chdir(str(resource_path_root / "pegasus-integrity"))
    yield
    os.chdir(cwd)


@pytest.mark.parametrize(
    "args,expected",
    [
        ("--generate=data.1", 0),
        ("--generate=data.1:data.2", 0),
        ("--generate-xml=data.1", 0),
        ("--generate-xml=foo1=data.1:foo2=data.2", 0),
    ],
)
def test_generate(data_files, args, expected):
    rv = subprocess.run([PYTHON_INTERPRETER, PEGASUS_INTEGRITY, args])

    assert rv.returncode == expected


@pytest.mark.parametrize(
    "args,expected", [("--verify=data.1", 0), ("--verify=data.1:foo.2=data.2", 0),],
)
def test_verify(data_files, args, expected):
    rv = subprocess.run([PYTHON_INTERPRETER, PEGASUS_INTEGRITY, args])

    assert rv.returncode == expected


def test_verify_multiple_stdin(data_files):
    print(cli.__path__)
    rv = subprocess.run(
        [PYTHON_INTERPRETER, PEGASUS_INTEGRITY, "--print-timings", "--verify=stdin"],
        input=b"data.1:foo.2=data.2",
    )

    assert rv.returncode == 0
